package jp.ne.xuni.selene.locators;

import java.util.ArrayList;

import org.eclipse.jface.dialogs.IDialogSettings;
import org.eclipse.swt.widgets.Composite;

import jp.ne.xuni.selene.Query;
import jp.ne.xuni.selene.SearchResult;
import jp.ne.xuni.selene.SelenePlugin;
import jp.ne.xuni.selene.Snippet;

public abstract class SnippetLocator {
	private static final String DS_ID = ".locators";
	private static final String SEPARATOR = "[^a-z]";

	protected int debugLevel = 0;

	protected IDialogSettings ds = SelenePlugin.getDefault().
		getDialogSettings().getSection(SelenePlugin.PLUGIN_ID + DS_ID);
	{
		if (ds == null) {
			ds = SelenePlugin.getDefault().getDialogSettings().addNewSection(
				SelenePlugin.PLUGIN_ID + DS_ID);
		}
		try {
			debugLevel = ds.getInt("debugLevel");
		} catch (NumberFormatException e) { }
	}

	public SnippetLocator() {
	}

	/*
	 * Methods overriding this method must be thread safe.
	 */
	public abstract int locate(
		Query query, SearchResult result, Snippet snippet);

	/*
	 * This method will be called once when SeleneView initializes UI
	 * components.
	 */
	public void setup(Composite panel) {
	}

	public static String[][] splitDocument(String document) {
		return splitDocument(document, SEPARATOR, 1, 0);
	}
	public static String[][] splitDocument(String document,
		String separator) {
		return splitDocument(document, separator, 1, 0);
	}
	public static String[][] splitDocument(String document,
		int size, int overlap) {
		return splitDocument(document, SEPARATOR, size, overlap);
	}
	public static String[][] splitDocument(String document,
		String separator, int size, int overlap) {
		String[] lines = document.split(Snippet.DELIMITER);
		lines = concatLines(lines, size, Math.max(size - overlap, 1));

		String[][] words = new String[lines.length][];
		for (int i = 0; i < lines.length; i++) {
			words[i] = splitLine(lines[i], separator);
		}
		return words;
	}
	private static String[] concatLines(String[] lines, int size, int each) {
		if (size <= 1) return lines;

		String[] blocks = new String[(lines.length + each - 1) / each];
		for (int i = 0, start = 0; start < lines.length; i++) {
			blocks[i] = "";
			for (int j = start; j < start + size && j < lines.length; j++) {
				blocks[i] += lines[j] + Snippet.DELIMITER;
			}
			start += each;
		}
		return blocks;
	}

	public static String[] splitLine(String line) {
		return splitLine(line, SEPARATOR);
	}
	public static String[] splitLine(String line, String separator) {
		return filter(line.toLowerCase().split(separator));
	}
	private static String[] filter(String[] words) {
		ArrayList<String> wordList = new ArrayList<String>(words.length);
		for (String word : words) {
			if (word.length() > 0) wordList.add(word);
		}
		return wordList.toArray(new String[wordList.size()]);
	}
}
