package jp.ne.xuni.selene.locators;

import java.io.FileNotFoundException;
import java.io.FileReader;
import java.io.IOException;

import org.eclipse.swt.SWT;
import org.eclipse.swt.events.SelectionAdapter;
import org.eclipse.swt.events.SelectionEvent;
import org.eclipse.swt.events.SelectionListener;
import org.eclipse.swt.layout.GridData;
import org.eclipse.swt.layout.GridLayout;
import org.eclipse.swt.widgets.Button;
import org.eclipse.swt.widgets.Composite;
import org.eclipse.swt.widgets.Group;
import org.eclipse.swt.widgets.Label;
import org.eclipse.swt.widgets.Slider;

import jp.ne.xuni.selene.Index;
import jp.ne.xuni.selene.Query;
import jp.ne.xuni.selene.SearchResult;
import jp.ne.xuni.selene.Snippet;

// changed algorithm greatly at Revision 1.5.

public class TokenDensityLocator extends SnippetLocator {
	// // members for score propagation
	// // these members may be altered while running through control panels.
	// propagation flag
	protected boolean m_disablePropagation = false;
	static protected String m_DSDisablePropagation = "TDDisable";
	// window width (must be >= 0)
	protected int m_scoreWidth;
	static protected int m_DefaultScoreWidth = 5;
	static protected String m_DSScoreWidth = "TDScoreWidth";
	// function type
	static final protected int FT_DELTA = 0;
	static final protected int FT_LINEAR = 1;
	static final protected int FT_INVQUAD = 2;
	protected int m_funcType;
	static protected int m_DefaultFuncType = FT_INVQUAD;
	static protected String m_DSFuncType = "TDFuncType";
	// coefficient for each function
	protected double m_coeffLinear;
	static protected double m_DefaultCoeffLinear = 0.2;
	static protected String m_DSCoeffLinear = "TDCoeffLinear";
	protected double[] m_coeffInvQuad;
	static protected double[] m_DefaultCoeffInvQuad = { 1.0, 2.0, 1.0 };
	static protected String[] m_DSCoeffInvQuad = 
	{ "TDCoeffInvQuad0", "TDCoeffInvQuad1", "TDCoeffInvQuad2" };

	// allow console output (for debugging)
	// this member cannot(should not) be altered while running
	static protected boolean m_AllowDebugOutput = false;

	// this method must be thread-safe(reentrant)
	public int locate(Query query, SearchResult result, Snippet snippet) {
		return findPos(query, snippet.getDocument());
	}

	public int findPos(Query query, String target) {
		// obtain values for propagation from members
		boolean disablePropagation = m_disablePropagation;
		int scoreWidth = m_scoreWidth;
		int funcType = m_funcType;
		double[] coeff = new double[3];
		if ( ! disablePropagation ) {
			switch ( funcType ) {
			case FT_DELTA:
				break;
			case FT_LINEAR:
				coeff[0] = m_coeffLinear;
				break;
			case FT_INVQUAD:
				coeff[0] = m_coeffInvQuad[0];
				coeff[1] = m_coeffInvQuad[1];
				coeff[2] = m_coeffInvQuad[2];
				break;
			default:
				;
			}
		}

		if ( m_AllowDebugOutput ) {
			System.out.println("TokenDensityLocator.locate CP:"
					+ query.getCursorPosition()
					+ " / AC:" + query.getAttCoeff()
					+ " / TH:" + query.getThreshold()
					+ " / BS:" + query.getBlockSize()
					+ " / BO:" + query.getBlockOverlap());
			if ( ! disablePropagation ) {
				switch ( funcType ) {
				case FT_DELTA:
					System.out.println("TokenDensityLocator.locate SW:"
							+ scoreWidth + " / FT:DELTA");
					break;
				case FT_LINEAR:
					System.out.println("TokenDensityLocator.locate SW:"
							+ scoreWidth + " / FT:LINEAR " + coeff[0]);
					break;
				case FT_INVQUAD:
					System.out.println("TokenDensityLocator.locate SW:"
							+ scoreWidth + " / FT:INVQUAD "
							+ coeff[0] + " " + coeff[1] + " " + coeff[2]);
					break;
				}
			} else {
				System.out.println("TokenDensityLocator.locate NO PROPAGATION");
			}
		}

		// get Index data
		Index queryIndex = query.getIndex();
		// split target text.
		String[][] targetWordArray =
			splitDocument(target, query.getBlockSize(), query.getBlockOverlap());
		// for test use: check the result of splitter
		if ( m_AllowDebugOutput ) {
			/*
			for ( int i = 0 ; i < targetWordArray.length ; ++i ) {
				if ( targetWordArray[i].length > 0 ) {
					System.out.print("T" + i + ":");
					for ( int j = 0 ; j < targetWordArray[i].length ; ++j ) {
						System.out.print(" " + targetWordArray[i][j]);
					}
					System.out.println("");
				}
			}
			*/
		}

		// initialize score array
		double[] targetScoreArray = new double[targetWordArray.length];
		for ( int i = 0 ; i < targetWordArray.length ; ++i ) {
			targetScoreArray[i] = 0.0;
		}

		// add scores by each word of target, considering TF&DF.
		for ( int i = 0 ; i < targetWordArray.length ; ++i ) {
			for ( int j = 0 ; j < targetWordArray[i].length ; ++j ) {
				int tf = queryIndex.getTf(targetWordArray[i][j]);
				int df = queryIndex.getDf(targetWordArray[i][j]);
				if ( tf > 0 && df > 0 ) {
					double wordScore = tf / Math.log(df);
					if ( wordScore > 0.0 ) {
						if ( ! disablePropagation ) {
							distributeScore(targetScoreArray, i, wordScore,
									scoreWidth, funcType, coeff);
						} else {
							targetScoreArray[i] += wordScore;
						}
					}
					if ( m_AllowDebugOutput ) {
						System.out.println(targetWordArray[i][j] + ": TF" + tf + " DF" + df);
					}
				}
			}
		}

		// detect max score
		int result = 0;
		double maxScore = -Double.MAX_VALUE;
		for ( int i = 0 ; i < targetScoreArray.length ; ++i ) {
			if ( targetScoreArray[i] > maxScore ) {
				result = i;
				maxScore = targetScoreArray[i];
			}
		}
		// for test use:
		if ( m_AllowDebugOutput ) {
			/*
			for ( int i = 0 ; i < targetScoreArray.length ; ++i ) {
				if ( targetScoreArray[i] > (maxScore * 0.01) ) {
					System.out.println(i + ": " + targetScoreArray[i]);
				}
			}
			System.out.println("Result: " + result + "/" + targetScoreArray.length +
					" (" + maxScore + ")");
			*/
		}

		return result * Math.max(query.getBlockSize() - query.getBlockOverlap(), 1);
	}

	public double weightFuncInvQuad(int distance, double k0, double k1, double k2) {
		double r = Math.abs(distance);
		return 1.0 / (k0 + k1*r + k2*r*r);
	}
	public double weightFuncInvQuad(int distance, double coeff[]) {
		return weightFuncInvQuad(distance, coeff[0], coeff[1], coeff[2]);
	}
	public double weightFuncLinear(int distance, double decay) {
		return 1.0 - Math.abs(distance) * decay;
	}
	public double weightFuncDelta(int distance) {
		return (distance == 0) ? 1.0 : 0.0;
	}

	public void distributeScore(double[] array, int index, double score,
			int width, int funcType, double[] coeff) {
		for ( int d = -width ; d <= width ; ++d ) {
			int u = index + d;
			if ( 0 <= u && u < array.length ) {
				double weight = 0.0;
				switch ( funcType ) {
				case FT_DELTA:
					weight = weightFuncDelta(d);
					break;
				case FT_LINEAR:
					weight = weightFuncLinear(d, coeff[0]);
					break;
				case FT_INVQUAD:
					weight = weightFuncInvQuad(d, coeff);
					break;
				default:
				}
				if ( weight > 0.0 ) {
					array[u] += score * weight;
				}
			}
		}
	}

	// constructor
	public TokenDensityLocator() {
		m_AllowDebugOutput = (debugLevel >= 1) ? true : false;
		setEnablePropagation();
		setScoreWidth();
		setScorePropagation();
	}
	protected void setEnablePropagation() {
		// propagation flag will be turned off only when explicit "NO".
		m_disablePropagation = ds.getBoolean(m_DSDisablePropagation);
	}
	protected void setScoreWidth() {
		int w = m_DefaultScoreWidth; // first default
		try {
			w = ds.getInt(m_DSScoreWidth);
		} catch ( NumberFormatException ex ) {}
		m_scoreWidth = (w >= 0) ? w : 0;
		if ( m_AllowDebugOutput ) {
			System.out.println("TokenDensityLocator.setScoreWidth " + m_scoreWidth);
		}
	}
	protected void setScorePropagation() {
		// m_funcType
		int ft = m_DefaultFuncType; // first default
		try {
			ft = ds.getInt(m_DSFuncType);
		} catch ( NumberFormatException ex ) {}
		m_funcType = ft;
		// m_coeffLinear
		double cl = m_DefaultCoeffLinear;
		try {
			cl = ds.getDouble(m_DSCoeffLinear);
		} catch ( NumberFormatException ex ) {}
		m_coeffLinear = cl;
		// m_coeffInvQuad;
		double[] ciq = new double[3];
		for ( int i = 0 ; i < 3 ; ++i ) {
			ciq[i] = m_DefaultCoeffInvQuad[i];
			try {
				ciq[i] = ds.getDouble(m_DSCoeffInvQuad[i]);
			} catch ( NumberFormatException ex ) {}
		}
		m_coeffInvQuad = ciq;
		// debug output and error check
		switch ( m_funcType ) {
		case FT_DELTA:
		case FT_LINEAR:
		case FT_INVQUAD:
			if ( m_AllowDebugOutput ) {
				System.out.println("TokenDensityLocator.setScorePropagation: "
						+ m_funcType + " / "
						+ m_coeffLinear + " / "
						+ m_coeffInvQuad[0] + " " + m_coeffInvQuad[1] + " " + m_coeffInvQuad[2]);
			}
			break;
		default:
			System.err.println("TokenDensityLocator.setScorePropagation: illegal funcType: " + m_funcType);
		}
	}

	static public String trimString(String str, int size) {
		return (str.length() > size) ? str.substring(0, size) : str;
	}

	// panel setup
	// variables which are accessed from SelectionLister
	protected Button bDisable;
	protected Slider sWidth;
	protected Label lWidthNow;
	protected Button[] bFunc;
	protected Group gCLinear;
	protected Slider sCLinear;
	protected Label lCLinearNow;
	protected Group gCInvQuad;
	protected Slider[] sCInvQuad;
	protected Label[] lCInvQuadNow;
	public void setup(Composite panel) {
		// make main group
		Group group = new Group(panel, SWT.NONE);
		group.setText("TokenDensityLocator");
		GridData griddata = new GridData(GridData.HORIZONTAL_ALIGN_BEGINNING);
		griddata.horizontalSpan = 3;
		group.setLayoutData(griddata);
		GridLayout layout = new GridLayout();
		layout.numColumns = 1;
		group.setLayout(layout);
		// propagation disable flag
		bDisable = new Button(group, SWT.CHECK);
		bDisable.setText("Disable Propagation");
		bDisable.setSelection(m_disablePropagation);
		bDisable.addSelectionListener(new SelectionAdapter() {
			public void widgetSelected(SelectionEvent e) {
				boolean b = bDisable.getSelection();
				if ( b != m_disablePropagation ) {
					m_disablePropagation = b;
					ds.put(m_DSDisablePropagation, m_disablePropagation);
				}
			}
		});
		// propagation width
		Group gWidth = new Group(group, SWT.NONE);
		gWidth.setText("Propagation Width:");
		GridLayout layWidth = new GridLayout();
		layWidth.numColumns = 2;
		gWidth.setLayout(layWidth);
		sWidth = new Slider(gWidth, SWT.HORIZONTAL);
		sWidth.setValues(m_scoreWidth, 1, 20+2, 2, 1, 3);
		lWidthNow = new Label(gWidth, SWT.NONE);
		lWidthNow.setText(trimString(String.valueOf(m_scoreWidth) + "  ", 3));
		sWidth.addSelectionListener(new SelectionAdapter() {
			public void widgetSelected(SelectionEvent e) {
				int width = sWidth.getSelection();
				if ( width != m_scoreWidth ) {
					m_scoreWidth = width;
					ds.put(m_DSScoreWidth, m_scoreWidth);
					lWidthNow.setText(String.valueOf(m_scoreWidth));
				}
			}
		});
		// propagation function
		Group gFunc = new Group(group, SWT.NONE);
		gFunc.setText("Function Type:");
		GridLayout layFunc = new GridLayout();
		layFunc.numColumns = 3;
		gFunc.setLayout(layFunc);
		bFunc = new Button[3];
		SelectionListener slFunc = new SelectionAdapter() {
			public void widgetSelected(SelectionEvent e) {
				int ft = m_funcType;
				if ( bFunc[0].getSelection() ) {
					ft = FT_DELTA;
				} else if ( bFunc[1].getSelection() ) {
					ft = FT_LINEAR;
				} else if ( bFunc[2].getSelection() ) {
					ft = FT_INVQUAD;
				}
				if ( ft != m_funcType ) {
					m_funcType = ft;
					ds.put(m_DSFuncType, m_funcType);
					//gCLinear.setEnabled(m_funcType == FT_LINEAR);
					//gCInvQuad.setEnabled(m_funcType == FT_INVQUAD);
				}
			}
		};
		for ( int i = 0 ; i < bFunc.length ; ++i ) {
			bFunc[i] = new Button(gFunc, SWT.RADIO);
			bFunc[i].addSelectionListener(slFunc);
		}
		bFunc[0].setText("Delta");
		bFunc[1].setText("Linear");
		bFunc[2].setText("InvQuad");
		bFunc[0].setSelection(m_funcType == FT_DELTA);
		bFunc[1].setSelection(m_funcType == FT_LINEAR);
		bFunc[2].setSelection(m_funcType == FT_INVQUAD);
		// propagation coeff linear
		gCLinear = new Group(group, SWT.NONE);
		gCLinear.setText("Coefficient for Linear:");
		GridLayout layCLinear = new GridLayout();
		layCLinear.numColumns = 2;
		gCLinear.setLayout(layCLinear);
		sCLinear = new Slider(gCLinear, SWT.HORIZONTAL);
		sCLinear.setValues((int)(m_coeffLinear * 100), 0, 200+20, 20, 1, 10);
		lCLinearNow = new Label(gCLinear, SWT.NONE);
		lCLinearNow.setText(trimString(String.valueOf(m_coeffLinear) + "   ", 5));
		sCLinear.addSelectionListener(new SelectionAdapter() {
			public void widgetSelected(SelectionEvent e) {
				double coeff = (double)sCLinear.getSelection() * 0.01;
				if ( coeff != m_coeffLinear ) {
					m_coeffLinear = coeff;
					ds.put(m_DSCoeffLinear, m_coeffLinear);
					lCLinearNow.setText(trimString(String.valueOf(m_coeffLinear), 5));
				}
			}
		});
		//gCLinear.setEnabled(m_funcType == FT_LINEAR);
		// propagation coeff invquad
		gCInvQuad = new Group(group, SWT.NONE);
		gCInvQuad.setText("Coefficients for InvQuad:");
		GridLayout layCInvQuad = new GridLayout();
		layCInvQuad.numColumns = 3;
		gCInvQuad.setLayout(layCInvQuad);
		Label[] lCInvQuad = new Label[3];
		sCInvQuad = new Slider[3];
		lCInvQuadNow = new Label[3];
		SelectionListener slCInvQuad = new SelectionAdapter() {
			public void widgetSelected(SelectionEvent e) {
				for ( int i = 0 ; i < sCInvQuad.length ; ++i ) {
					double coeff = sCInvQuad[i].getSelection() * 0.01;
					if ( coeff != m_coeffInvQuad[i] ) {
						m_coeffInvQuad[i] = coeff;
						ds.put(m_DSCoeffInvQuad[i], m_coeffInvQuad[i]);
						lCInvQuadNow[i].setText(trimString(String.valueOf(m_coeffInvQuad[i]), 5));
					}
				}
			}
		};
		for ( int i = 0 ; i < lCInvQuad.length ; ++i ) {
			lCInvQuad[i] = new Label(gCInvQuad, SWT.NONE);
			sCInvQuad[i] = new Slider(gCInvQuad, SWT.HORIZONTAL);
			sCInvQuad[i].setValues((int)(m_coeffInvQuad[i] * 100), 0, 400+20, 20, 1, 10);
			sCInvQuad[i].addSelectionListener(slCInvQuad);
			lCInvQuadNow[i] = new Label(gCInvQuad, SWT.NONE);
			lCInvQuadNow[i].setText(trimString(String.valueOf(m_coeffInvQuad[i]) + "   ", 5));
		}
		lCInvQuad[0].setText("const");
		lCInvQuad[1].setText("linear");
		lCInvQuad[2].setText("quad");
		//gCInvQuad.setEnabled(m_funcType == FT_INVQUAD);
	}

	// !! never works after Revision 1.5 !!
	// static test methods
	static protected String readText(String path, int pos, int size) {
		String result = null;
		try {
			FileReader fr = new FileReader(path);
			char[] cbuf = new char[size + 1];
			int n = fr.read(cbuf, pos, size);
			if ( n > 0 ) { // read correctly
				result = new String(cbuf, 0, n);
			}
			fr.close();
		} catch ( FileNotFoundException ex ) {
		} catch ( IOException ex ) {
		}
		return result;
	}
	static public void main(String[] args) {
		TokenDensityLocator fpt = new TokenDensityLocator();
		String query = "a\nb";
		String target = "z\na\nb\nc\nb a\na";
		{
			String s;
			s = readText("./TokenDensityLocator.java", 0, 127);
			if ( s != null ) {
				query = s;
			}
			s = readText("./TokenDensityLocator.java", 0, 1023);
			if ( s != null ) {
				target = s;
			}
		}
		Query q = new Query();
		q.setDocument(query);
		int result = fpt.findPos(q, target);
		System.out.println("Result: " + result);
	}
}
