package jp.ne.xuni.selene.views;

import jp.ne.xuni.selene.DirectSearchResult;
import jp.ne.xuni.selene.FetchedDocument;
import jp.ne.xuni.selene.Query;
import jp.ne.xuni.selene.SearchResult;
import jp.ne.xuni.selene.SeleneClient;
import jp.ne.xuni.selene.SelenePlugin;
import jp.ne.xuni.selene.Snippet;
import jp.ne.xuni.selene.ZippedSearchResult;
import jp.ne.xuni.selene.SeleneClient.FetchDocument;
import jp.ne.xuni.selene.SeleneClient.Search;
import jp.ne.xuni.selene.SeleneClient.SearchSnippet;
import jp.ne.xuni.selene.locators.SnippetLocator;
import jp.ne.xuni.seleneserver.ISeleneServer;

import org.eclipse.core.runtime.IProgressMonitor;
import org.eclipse.core.runtime.IStatus;
import org.eclipse.core.runtime.Status;
import org.eclipse.core.runtime.jobs.IJobChangeEvent;
import org.eclipse.core.runtime.jobs.Job;
import org.eclipse.core.runtime.jobs.JobChangeAdapter;
import org.eclipse.jface.action.Action;
import org.eclipse.jface.action.IAction;
import org.eclipse.jface.action.IMenuListener;
import org.eclipse.jface.action.IMenuManager;
import org.eclipse.jface.action.IToolBarManager;
import org.eclipse.jface.action.MenuManager;
import org.eclipse.jface.action.Separator;
import org.eclipse.jface.commands.ActionHandler;
import org.eclipse.jface.dialogs.IDialogSettings;
import org.eclipse.jface.text.DocumentEvent;
import org.eclipse.jface.text.IDocument;
import org.eclipse.jface.text.IDocumentListener;
import org.eclipse.jface.text.IDocumentPartitioningListener;
import org.eclipse.jface.text.TextSelection;
import org.eclipse.jface.viewers.ColumnWeightData;
import org.eclipse.jface.viewers.DoubleClickEvent;
import org.eclipse.jface.viewers.IDoubleClickListener;
import org.eclipse.jface.viewers.ISelection;
import org.eclipse.jface.viewers.ISelectionProvider;
import org.eclipse.jface.viewers.IStructuredContentProvider;
import org.eclipse.jface.viewers.IStructuredSelection;
import org.eclipse.jface.viewers.ITableLabelProvider;
import org.eclipse.jface.viewers.LabelProvider;
import org.eclipse.jface.viewers.TableLayout;
import org.eclipse.jface.viewers.TableViewer;
import org.eclipse.jface.viewers.Viewer;
import org.eclipse.swt.SWT;
import org.eclipse.swt.custom.ScrolledComposite;
import org.eclipse.swt.custom.StackLayout;
import org.eclipse.swt.events.ModifyEvent;
import org.eclipse.swt.events.ModifyListener;
import org.eclipse.swt.events.SelectionAdapter;
import org.eclipse.swt.events.SelectionEvent;
import org.eclipse.swt.events.SelectionListener;
import org.eclipse.swt.graphics.Color;
import org.eclipse.swt.graphics.Image;
import org.eclipse.swt.layout.GridData;
import org.eclipse.swt.layout.GridLayout;
import org.eclipse.swt.widgets.Button;
import org.eclipse.swt.widgets.Composite;
import org.eclipse.swt.widgets.Control;
import org.eclipse.swt.widgets.Display;
import org.eclipse.swt.widgets.Group;
import org.eclipse.swt.widgets.Label;
import org.eclipse.swt.widgets.Menu;
import org.eclipse.swt.widgets.Slider;
import org.eclipse.swt.widgets.Spinner;
import org.eclipse.swt.widgets.Table;
import org.eclipse.swt.widgets.TableColumn;
import org.eclipse.swt.widgets.Text;
import org.eclipse.ui.IActionBars;
import org.eclipse.ui.IEditorInput;
import org.eclipse.ui.IEditorPart;
import org.eclipse.ui.IEditorReference;
import org.eclipse.ui.IMemento;
import org.eclipse.ui.IPartListener;
import org.eclipse.ui.IReusableEditor;
import org.eclipse.ui.ISharedImages;
import org.eclipse.ui.IViewSite;
import org.eclipse.ui.IWorkbenchActionConstants;
import org.eclipse.ui.IWorkbenchPage;
import org.eclipse.ui.IWorkbenchPart;
import org.eclipse.ui.PartInitException;
import org.eclipse.ui.PlatformUI;
import org.eclipse.ui.handlers.IHandlerService;
import org.eclipse.ui.part.ViewPart;
import org.eclipse.ui.progress.IWorkbenchSiteProgressService;
import org.eclipse.ui.texteditor.ITextEditor;

/**
 * This sample class demonstrates how to plug-in a new
 * workbench view. The view shows data obtained from the
 * model. The sample creates a dummy model on the fly,
 * but a real implementation would connect to the model
 * available either in this or another plug-in (e.g. the workspace).
 * The view is connected to the model using a content provider.
 * <p>
 * The view uses a label provider to define how model
 * objects should be presented in the view. Each
 * view can present the same model objects using
 * different labels and icons, if needed. Alternatively,
 * a single label provider can be shared between views
 * in order to ensure that objects of the same type are
 * presented in the same way everywhere.
 * <p>
 */

public class SeleneView extends ViewPart implements IPartListener {
	private static final String DS_ID = ".view";
	private static final String ERR_MSG = "SeleneView error: ";
	private static final int MAX_SNIPPETS = 8;
	private static final String[] HEADINGS =
		{ "Score", "Document Name", "Line" };
	// To introduce a new algorithm, change this and the next String array,
	// which represents class names to be loaded.
	private static final String[] ALGORITHMS =
		{ "Set size", "Split index", "Token density", "Alignment" };
	private static final String[] LOCATOR_CLASSES =
		{ "DirectSearchLocator", "TokenDensityLocator", "AlignmentLocator" };
	private static final String LOCATOR_PACKAGE =
		SelenePlugin.PLUGIN_ID + ".locators";
	private static final FetchedEditorInput defaultEditorInput =
		new FetchedEditorInput(new Snippet("", new SearchResult()));

	private Composite parent;
	private StackLayout layout;
	private Control[] panels = new Control[2];
	private Group group0;
	private Label label0_0;
	private Label label0_1;
	private Label label0_2;
	private Slider slider0_0;
	private Group group1;
	private Slider slider1_0;
	private Label label1_0;
	private Button buttonLineWeighted;
	private Group group2;
	private Job job2;
	private Label label2_0;
	private Label label2_1;
	private Text text2_0;
	private Text text2_1;
	private Group group3;
	private Button[] button3 = new Button[ALGORITHMS.length];
	private Group group4;
	private Label label4_0;
	private Slider slider4_0;
	private Group group5;
	private Label label5_0;
	private Label label5_1;
	private Spinner spinner5_0;
	private Spinner spinner5_1;

	private TableViewer viewer;
	private IWorkbenchSiteProgressService ps;
	private IWorkbenchPage page;
	private ITextEditor editor;
	private IDocument document;
	private Action stopSearch;
	private Action searchAction;
	private Action fetchAction;
	private Action showControlPanel;
	private Runnable searcher = new Runnable() {
		public void run() {
			doSearch();
		}
	};
	private IMemento memento;

	private SeleneClient client = new SeleneClient();
	private Search job = client.new Search("Selene Search");
	private SearchResult[] result = new SearchResult[0];
	private double attCoeff = Query.DEFAULT_ATT_COEFF;
	private double threshold = Query.DEFAULT_THRESHOLD;
	private boolean lineWeighted;
	private int blockSize = Query.DEFAULT_BLOCK_SIZE;
	private int blockOverlap = 0;
	private IEditorPart[] snippetEditors = new IEditorPart[MAX_SNIPPETS];
	private Boolean searchTemporarilyDisabled = false;
	private boolean autoSearch;
	private int algorithm = 0;
	private SnippetLocator[] locators =
		new SnippetLocator[LOCATOR_CLASSES.length];
	{
		int i = 0;
		for (String locatorClass : LOCATOR_CLASSES) {
			try {
				locators[i] = (SnippetLocator)Class.forName(
					LOCATOR_PACKAGE + "." + locatorClass).newInstance();
			} catch (Exception e) {
				e.printStackTrace();
			}
			i++;
		}
	}
	int numOfSnippets = 3;

	private IDialogSettings ds = SelenePlugin.getDefault().
		getDialogSettings().getSection(SelenePlugin.PLUGIN_ID + DS_ID);
	{
		if (ds == null) {
			ds = SelenePlugin.getDefault().getDialogSettings().addNewSection(
				SelenePlugin.PLUGIN_ID + DS_ID);
		}
		String host = ds.get("host");
		if (host != null) client.setHost(host);
		String repository = ds.get("repository");
		if (repository != null) client.setRepository(repository);
		try {
			attCoeff = ds.getDouble("attCoeff");
		} catch (NumberFormatException e) { }
		try {
			threshold = ds.getDouble("threshold");
		} catch (NumberFormatException e) { }
		try {
			blockSize = ds.getInt("blockSize");
		} catch (NumberFormatException e) { }
		try {
			blockOverlap = ds.getInt("blockOverlap");
		} catch (NumberFormatException e) { }
		try {
			algorithm = ds.getInt("algorithm");
		} catch (NumberFormatException e) { }
		try {
			numOfSnippets = ds.getInt("numOfSnippets");
		} catch (NumberFormatException e) { }
		lineWeighted = ds.getBoolean("lineWeighted");
		autoSearch = ds.getBoolean("autoSearch");
	}

	private TimeCounter tc = new TimeCounter(this); // for time recording

	/*
	 * The content provider class is responsible for
	 * providing objects to the view. It can wrap
	 * existing objects in adapters or simply return
	 * objects as-is. These objects may be sensitive
	 * to the current input of the view, or ignore
	 * it and always show the same content 
	 * (like Task List, for example).
	 */

	class ViewContentProvider implements IStructuredContentProvider,
		IDocumentListener, IDocumentPartitioningListener {
		public void dispose() {
			if (document != null) {
				document.removeDocumentListener(this);
				document.removeDocumentPartitioningListener(this);
			}
		}

		public void inputChanged(Viewer v, Object oldInput, Object newInput) {
			if (oldInput != newInput) {
				if (document != null) {
					document.removeDocumentListener(this);
					document.removeDocumentPartitioningListener(this);
				}

				document = (IDocument)newInput;
				if (document != null) {
					document.addDocumentListener(this);
					document.addDocumentPartitioningListener(this);
					if (!searchTemporarilyDisabled && autoSearch) {
						launchJob(job);
					}
				}
			}
		}

		public Object[] getElements(Object parent) {
			return result;
		}

		public void documentAboutToBeChanged(DocumentEvent event) {
		}
		public void documentChanged(DocumentEvent event) {
		}

		public void documentPartitioningChanged(IDocument document) {
			if (autoSearch) {
				launchJob(job);
			}
		}
	}

	class ViewLabelProvider
		extends LabelProvider implements ITableLabelProvider {
		public Image getImage(Object obj) {
			return PlatformUI.getWorkbench().
				getSharedImages().getImage(ISharedImages.IMG_OBJ_FILE);
		}

		public String getText(Object obj) {
			return getColumnText(obj, 0) +
				ISeleneServer.SEPARATOR + getColumnText(obj, 1) +
				(obj instanceof DirectSearchResult ?
				ISeleneServer.SEPARATOR + getColumnText(obj, 2) : "");
		}

		public Image getColumnImage(Object obj, int index) {
			switch (index) {
				case 0:
					return null;
				case 1:
					return getImage(obj);
				default:
					return null;
			}
		}

		public String getColumnText(Object obj, int index) {
			SearchResult result = (SearchResult)obj;
			switch (index) {
				case 0:
					return String.valueOf(result.getWeight());
				case 1:
					return result.getName();
				default:
					int line = 0;
					if (result instanceof DirectSearchResult) {
						line = ((DirectSearchResult)result).getLine();
					}
					return line == 0 ? null : String.valueOf(line);
			}
		}
	}

	class SearchJobListener extends JobChangeAdapter {
		public void done(IJobChangeEvent event) {
			IStatus status = event.getResult();
			if (status.isOK()) {
				tc.recordTime(); // for time recording
				Search search = (Search)event.getJob();
				result = search.getSearchResult();

				for (int i = 0; i < numOfSnippets && i < result.length; i++) {
					SearchSnippet job;
					String jobName = "Selene Snippet Search (" +
						ALGORITHMS[algorithm] + ")";
					switch(algorithm) {
						case 0:
							job = client.new SearchSnippet(jobName, i, tc);
							break;
						default:
							job = client.new FetchSnippet(jobName, i, tc,
								locators[algorithm - 1], search.getQuery());
					}
					job.addJobChangeListener(new SnippetJobListener());
					job.setQuery(result[i]);
					job.setOriginalResult(result);
					launchJob(job);
				}
			} else {
				result = new SearchResult[0];
				if (status.matches(IStatus.ERROR)) stopSearch();
			}

			Control ctrl = viewer.getControl();
			if (ctrl != null && !ctrl.isDisposed()) {
				ctrl.getDisplay().asyncExec(new Runnable() {
					public void run() {
						Control ctrl = viewer.getControl();
						if (ctrl != null && !ctrl.isDisposed()) {
							viewer.refresh();
						}
					}
				});
			}
		}
	}

	class FetchJobListener extends JobChangeAdapter {
		public void done(IJobChangeEvent event) {
			if (event.getResult().isOK()) {
				openDocument(
					((FetchDocument)event.getJob()).getFetchResult());
			}
		}
	}

	class OpenDocument implements Runnable {
		protected FetchedDocument document;
		protected IEditorPart editor;

		public OpenDocument(FetchedDocument document) {
			this.document = document;
		}
		public OpenDocument(FetchedDocument document, IEditorPart editor) {
			this(document);
			this.editor = editor;
		}

		public void run() {
			openDocument(editor, false);
		}
		protected IEditorPart openDocument(
			IEditorPart editor, boolean isSnippet) {
			if (editor == null) {
				synchronized (searchTemporarilyDisabled) {
					if (isSnippet) searchTemporarilyDisabled = true;
					try {
						editor = page.openEditor(
							new FetchedEditorInput(document),
							PlatformUI.getWorkbench().getEditorRegistry().
							getDefaultEditor(document.getName()).getId(),
							true);
					} catch (PartInitException e) {
						System.err.println(ERR_MSG + e.getMessage());
					}
					if (isSnippet) searchTemporarilyDisabled = false;
				}
			} else if (editor instanceof IReusableEditor) {
				page.reuseEditor((IReusableEditor)editor,
					new FetchedEditorInput(document));
			} // TODO needs "else" case consideration
			if (editor instanceof ITextEditor &&
				document instanceof Snippet) {
				((ITextEditor)editor).getSelectionProvider().setSelection(
					new TextSelection(((Snippet)document).getCurrentOffset(),
					0));
			}
			return editor;
		}
	}

	class SnippetJobListener extends JobChangeAdapter {
		public void done(IJobChangeEvent event) {
			if (event.getResult().isOK()) {
				SearchSnippet job = (SearchSnippet)event.getJob();
				Control ctrl = viewer.getControl();
				if (ctrl != null && !ctrl.isDisposed()) {
					ctrl.getDisplay().asyncExec(new OpenSnippet(
						job.getSnippet(), job.getId()));
				}
			} else {
				stopSearch();
			}
		}
	}

	class OpenSnippet extends OpenDocument {
		protected int id;

		OpenSnippet(Snippet snippet, int id) {
			super(snippet);
			this.id = id;
		}

		public void run() {
			snippetEditors[id] = openDocument(snippetEditors[id], true);
			tc.recordTime(id); // for time recording
		}
	}

	public SeleneView() {
		super();
		job.addJobChangeListener(new SearchJobListener());
	}

	private void launchJob(Job job) {
		if (job instanceof Search) {
			Search search = (Search)job;
			search.setQuery(new Query(attCoeff, threshold, lineWeighted,
				blockSize, blockOverlap));
			if (document != null) {
				search.setQuery(document.get());
			}
			if (editor != null) {
				ISelectionProvider sp = editor.getSelectionProvider();
				if (sp != null) {
					TextSelection selection =
						(TextSelection)sp.getSelection();
					if (selection.getLength() == 0) {
						search.setCursorPosition(selection.getStartLine());
					} else {
						search.setQuery(selection.getText());
						search.setAttCoeff(Query.DEFAULT_ATT_COEFF);
					}
				}
			}
			tc.setStartTime(); // for time recording
		}

		ps.schedule(job);
	}

	public void stopSearch() {
		stopSearch.setChecked(true);
		stopSearch.run();
	}

	public void doSearch() {
		IEditorPart defaultEditor = page.findEditor(defaultEditorInput);
		if (defaultEditor != null) {
			partActivated(defaultEditor);
		}
		searchAction.run();
	}
	public void doSearch(int msec) {
		timerExec(msec, searcher);
	}
	void timerExec(int msec, Runnable runnable) {
		Control ctrl = viewer.getControl();
		if (ctrl != null && !ctrl.isDisposed()) {
			ctrl.getDisplay().timerExec(msec, runnable);
		}
	}

	Snippet[] getDocuments() {
		Snippet[] documents = new Snippet[numOfSnippets + 1];

		IEditorPart defaultEditor = page.findEditor(defaultEditorInput);
		if (defaultEditor != null) {
			documents[0] = (Snippet)((FetchedEditorInput)defaultEditor.
				getEditorInput()).getInput();
		} else if (document != null) {
			documents[0] = new Snippet(document.get(), new SearchResult());
			if (editor != null) {
				ISelectionProvider sp = editor.getSelectionProvider();
				if (sp != null) {
					documents[0].setCurrentLine(
						((TextSelection)sp.getSelection()).getStartLine());
				}
			}
		}

		for (int i = 0; i < numOfSnippets; i++) {
			if (snippetEditors[i] != null) {
				IEditorInput input = snippetEditors[i].getEditorInput();
				if (input instanceof FetchedEditorInput) {
					FetchedDocument document =
						((FetchedEditorInput)input).getInput();
					if (document instanceof Snippet) {
						documents[i + 1] = (Snippet)document;
					}
				}
			}
		}

		return documents;
	}

	public void fetchDocument(String name) {
		fetchDocument(new ZippedSearchResult(
			0.0, client.getHost(), client.getRepository(), name));
	}
	private void fetchDocument(SearchResult query) {
		FetchDocument job = client.new FetchDocument("Selene Fetch");
		job.addJobChangeListener(new FetchJobListener());
		job.setQuery(query);
		launchJob(job);
	}

	public void openDocument(String text) {
		openDocument(text, false);
	}
	public void openDocument(String text, boolean syncExec) {
		openDocument(text, syncExec, 0);
	}
	public void openDocument(
		String text, boolean syncExec, int cursorPosition) {
		Snippet document = new Snippet(text, new SearchResult());
		document.setCurrentLine(cursorPosition);
		openDocument(document, syncExec, page.findEditor(defaultEditorInput));
	}
	private void openDocument(FetchedDocument document) {
		openDocument(document, false);
	}
	private void openDocument(FetchedDocument document, boolean syncExec) {
		openDocument(document, syncExec, null);
	}
	private void openDocument(
		FetchedDocument document, boolean syncExec, IEditorPart editor) {
		Control ctrl = viewer.getControl();
		if (ctrl != null && !ctrl.isDisposed()) {
			Display display = ctrl.getDisplay();
			OpenDocument opener = new OpenDocument(document, editor);
			if (syncExec) {
				display.syncExec(opener);
			} else {
				display.asyncExec(opener);
			}
		}
	}

	/**
	 * This is a callback that will allow us
	 * to create the viewer and initialize it.
	 */
	public void createPartControl(Composite parent) {
		this.parent = parent;
		layout = new StackLayout();
		parent.setLayout(layout);

		panels[0] = makeTable();
		layout.topControl = panels[0];
		viewer = new TableViewer((Table)panels[0]);
		viewer.setContentProvider(new ViewContentProvider());
		viewer.setLabelProvider(new ViewLabelProvider());

		panels[1] = makeControlPanel();

		ps = (IWorkbenchSiteProgressService)getSite().
			getAdapter(IWorkbenchSiteProgressService.class);
		page = getSite().getPage();

		if (memento != null) {
			for (IEditorReference er : page.getEditorReferences()) {
				try {
					String path = er.getEditorInput().getToolTipText();
					for (int i = 0; i < MAX_SNIPPETS; i++) {
						if (path.equals(
							memento.getString("snippetEditor" + i))) {
							snippetEditors[i] = er.getEditor(true);
							break;
						}
					}
				} catch (PartInitException e) {
					System.err.println(ERR_MSG + e.getMessage());
				}
			}
			memento = null;
		}

		IEditorPart activeEditor = page.getActiveEditor();
		if (activeEditor instanceof ITextEditor) {
			editor = (ITextEditor)activeEditor;
			viewer.setInput(editor.
				getDocumentProvider().getDocument(editor.getEditorInput()));
		}
		page.addPartListener(this);

		makeActions();
		hookContextMenu();
		hookDoubleClickAction();
		contributeToActionBars();
	}

	private Table makeTable() {
		Table table = new Table(parent,
			SWT.H_SCROLL | SWT.V_SCROLL | SWT.SINGLE | SWT.FULL_SELECTION);
		TableLayout layout = new TableLayout();
		table.setLayout(layout);
		table.setLinesVisible(true);
		table.setHeaderVisible(true);

		layout.addColumnData(new ColumnWeightData(1, 40, true));
		TableColumn tc0 = new TableColumn(table, SWT.NONE);
		tc0.setText(HEADINGS[0]);
		tc0.setAlignment(SWT.LEFT);
		tc0.setResizable(true);

		layout.addColumnData(new ColumnWeightData(5, 200, true));
		TableColumn tc1 = new TableColumn(table, SWT.NONE);
		tc1.setText(HEADINGS[1]);
		tc1.setAlignment(SWT.LEFT);
		tc1.setResizable(true);

		layout.addColumnData(new ColumnWeightData(1, 40, true));
		TableColumn tc2 = new TableColumn(table, SWT.NONE);
		tc2.setText(HEADINGS[2]);
		tc2.setAlignment(SWT.LEFT);
		tc2.setResizable(true);

		return table;
	}

	private Composite makeControlPanel() {
		ScrolledComposite sc =
			new ScrolledComposite(parent, SWT.H_SCROLL | SWT.V_SCROLL);
		Composite panel = new Composite(sc, SWT.NONE);
		sc.setContent(panel);
		GridLayout layout = new GridLayout(3, false);
		layout.marginHeight = 10;
		layout.marginWidth = 10;
		panel.setLayout(layout);

		group0 = new Group(panel, SWT.NONE);
		group0.setText("Attenuation Coefficient");
		group0.setLayoutData(
			new GridData(SWT.BEGINNING, SWT.BEGINNING, false, false, 2, 1));
		group0.setLayout(new GridLayout(3, false));
		label0_0 = new Label(group0, SWT.NONE);
		label0_0.setText("Flat");
		slider0_0 = new Slider(group0, SWT.HORIZONTAL);
		slider0_0.setValues((int)(attCoeff * 100), 0, 210, 10, 1, 10);
		slider0_0.addSelectionListener(new SelectionAdapter() {
			public void widgetSelected(SelectionEvent e) {
				attCoeff = slider0_0.getSelection() / 100.0;
				ds.put("attCoeff", attCoeff);
				label0_2.setText(String.valueOf(attCoeff));
				group0.layout();
			}
		});
		label0_1 = new Label(group0, SWT.NONE);
		label0_1.setText("Steep");
		label0_2 = new Label(group0, SWT.NONE);
		label0_2.setText(String.valueOf(attCoeff));
		label0_2.setLayoutData(
			new GridData(SWT.CENTER, SWT.BEGINNING, false, false, 3, 1));

		group1 = new Group(panel, SWT.NONE);
		group1.setText("Rounding Threshold");
		group1.setLayoutData(
			new GridData(SWT.BEGINNING, SWT.BEGINNING, false, false));
		group1.setLayout(new GridLayout(1, false));
		slider1_0 = new Slider(group1, SWT.HORIZONTAL);
		slider1_0.setValues((int)(threshold * 100), 0, 110, 10, 1, 10);
		slider1_0.addSelectionListener(new SelectionAdapter() {
			public void widgetSelected(SelectionEvent e) {
				threshold = slider1_0.getSelection() / 100.0;
				ds.put("threshold", threshold);
				label1_0.setText(String.valueOf(threshold));
				group1.layout();
			}
		});
		label1_0 = new Label(group1, SWT.NONE);
		label1_0.setText(String.valueOf(threshold));
		label1_0.setLayoutData(
			new GridData(SWT.CENTER, SWT.BEGINNING, false, false));

		buttonLineWeighted = new Button(panel, SWT.CHECK);
		buttonLineWeighted.setText("Line weighted on 1st associative search");
		buttonLineWeighted.setLayoutData(
			new GridData(SWT.BEGINNING, SWT.BEGINNING, false, false, 3, 1));
		buttonLineWeighted.setSelection(lineWeighted);
		buttonLineWeighted.addSelectionListener(new SelectionAdapter() {
			public void widgetSelected(SelectionEvent e) {
				lineWeighted = buttonLineWeighted.getSelection();
				ds.put("lineWeighted", lineWeighted);
			}
		});

		group2 = new Group(panel, SWT.NONE);
		group2.setText("Server and Repository");
		group2.setLayoutData(
			new GridData(SWT.FILL, SWT.BEGINNING, false, false, 3, 1));
		group2.setLayout(new GridLayout(2, false));
		label2_0 = new Label(group2, SWT.NONE);
		label2_0.setText("Server name:");
		text2_0 = new Text(group2, SWT.BORDER | SWT.SINGLE);
		text2_0.setText(client.getHost());
		text2_0.setLayoutData(
			new GridData(SWT.FILL, SWT.CENTER, true, false));
		job2 = new Job("Server Lookup") {
			private Color colour0 = text2_0.getBackground();
			private Color colour1 =
				new Color(Display.getCurrent(), 255, 127, 127);

			private Runnable hostChecker = new Runnable() {
				public synchronized void run() {
					String host = text2_0.getText();
					ds.put("host", host);
					text2_0.setBackground(
						client.setHost(host) ? colour0 : colour1);
				}
			};

			protected IStatus run(IProgressMonitor monitor) {
				Control ctrl = viewer.getControl();
				if (ctrl != null && !ctrl.isDisposed()) {
					ctrl.getDisplay().asyncExec(hostChecker);
				}

				return new Status(IStatus.OK, SelenePlugin.PLUGIN_ID,
					"Server lookup finished");
			}
		};
		text2_0.addModifyListener(new ModifyListener() {
			public void modifyText(ModifyEvent e) {
				job2.schedule(500);
			}
		});
		label2_1 = new Label(group2, SWT.NONE);
		label2_1.setText("Repository:");
		text2_1 = new Text(group2, SWT.BORDER | SWT.SINGLE);
		text2_1.setText(client.getRepository());
		text2_1.setLayoutData(
			new GridData(SWT.FILL, SWT.CENTER, true, false));
		text2_1.addModifyListener(new ModifyListener() {
			public void modifyText(ModifyEvent e) {
				String repository = text2_1.getText();
				ds.put("repository", repository);
				client.setRepository(repository);
			}
		});

		group3 = new Group(panel, SWT.NONE);
		group3.setText("Snippet Search Algorithm");
		group3.setLayoutData(
			new GridData(SWT.BEGINNING, SWT.BEGINNING, false, false, 3, 1));
		group3.setLayout(new GridLayout(button3.length, false));
		SelectionListener listener = new SelectionAdapter() {
			public void widgetSelected(SelectionEvent e) {
				for (int i = 0; i < button3.length; i++) {
					if (button3[i].getSelection()) {
						algorithm = i;
						ds.put("algorithm", algorithm);
						break;
					}
				}
			}
		};
		for (int i = 0; i < button3.length; i++) {
			button3[i] = new Button(group3, SWT.RADIO);
			button3[i].setText(ALGORITHMS[i]);
			button3[i].addSelectionListener(listener);
		}
		button3[algorithm].setSelection(true);

		group4 = new Group(panel, SWT.NONE);
		group4.setText("Number of Snippets");
		group4.setLayoutData(
			new GridData(SWT.BEGINNING, SWT.BEGINNING, false, false));
		group4.setLayout(new GridLayout(1, false));
		slider4_0 = new Slider(group4, SWT.HORIZONTAL);
		slider4_0.setValues(numOfSnippets, 1, MAX_SNIPPETS + 1, 1, 1, 1);
		slider4_0.addSelectionListener(new SelectionAdapter() {
			public void widgetSelected(SelectionEvent e) {
				numOfSnippets = slider4_0.getSelection();
				ds.put("numOfSnippets", numOfSnippets);
				label4_0.setText(String.valueOf(numOfSnippets));
				group4.layout();
			}
		});
		label4_0 = new Label(group4, SWT.NONE);
		label4_0.setText(String.valueOf(numOfSnippets));
		label4_0.setLayoutData(
			new GridData(SWT.CENTER, SWT.BEGINNING, false, false));

		group5 = new Group(panel, SWT.NONE);
		group5.setText("Block Size and Overlap");
		group5.setLayoutData(
			new GridData(SWT.BEGINNING, SWT.BEGINNING, false, false, 2, 1));
		group5.setLayout(new GridLayout(4, true));
		label5_0 = new Label(group5, SWT.NONE);
		label5_0.setText("Size:");
		label5_0.setLayoutData(
			new GridData(SWT.END, SWT.CENTER, false, false));
		spinner5_0 = new Spinner(group5, SWT.NONE);
		spinner5_0.setValues(blockSize, 1, 99, 0, 1, 10);
		spinner5_0.addSelectionListener(new SelectionAdapter() {
			public void widgetSelected(SelectionEvent e) {
				blockSize = spinner5_0.getSelection();
				ds.put("blockSize", blockSize);
				if (blockOverlap >= blockSize) {
					blockOverlap = blockSize - 1;
					spinner5_1.setSelection(blockOverlap);
					ds.put("blockOverlap", blockOverlap);
				}
			}
		});
		label5_1 = new Label(group5, SWT.NONE);
		label5_1.setText("Overlap:");
		label5_1.setLayoutData(
			new GridData(SWT.END, SWT.CENTER, false, false));
		spinner5_1 = new Spinner(group5, SWT.NONE);
		spinner5_1.setValues(blockOverlap, 0, 98, 0, 1, 10);
		spinner5_1.addSelectionListener(new SelectionAdapter() {
			public void widgetSelected(SelectionEvent e) {
				blockOverlap = spinner5_1.getSelection();
				if (blockOverlap >= blockSize) {
					blockOverlap = blockSize - 1;
					spinner5_1.setSelection(blockOverlap);
				}
				ds.put("blockOverlap", blockOverlap);
			}
		});

		setupLocators(panel);

		panel.setSize(panel.computeSize(SWT.DEFAULT, SWT.DEFAULT));
		return sc;
	}
	private void setupLocators(Composite panel) {
		for (SnippetLocator locator : locators) {
			locator.setup(panel);
		}
	}

	private void makeActions() {
		stopSearch =
			new Action("Stop Automatic Search", IAction.AS_CHECK_BOX) {
				public void run() {
					if (this.isChecked()) {
						autoSearch = false;
						ds.put("autoSearch", false);
						stopSearch.setToolTipText("Start Automatic Search");
					} else {
						autoSearch = true;
						ds.put("autoSearch", true);
						stopSearch.setToolTipText("Stop Automatic Search");
					}
				}
			};
		stopSearch.setImageDescriptor(SeleneImages.STOP_SEARCH);
		if (!autoSearch) stopSearch.setChecked(true);
		stopSearch.run();

		searchAction = new Action("Search") {
			public void run() {
				launchJob(job);
			}
		};
		searchAction.setToolTipText("Search Documents");
		searchAction.setImageDescriptor(SeleneImages.SEARCH);
		((IHandlerService)PlatformUI.getWorkbench().
			getService(IHandlerService.class)).
			activateHandler(SelenePlugin.PLUGIN_ID + ".commands.search",
			new ActionHandler(searchAction));

		fetchAction = new Action("Document Fetch") {
			public void run() {
				ISelection selection = viewer.getSelection();
				Object obj =
					((IStructuredSelection)selection).getFirstElement();
				if (obj instanceof SearchResult) {
					fetchDocument((SearchResult)obj);
				}
			}
		};
		fetchAction.setToolTipText("Fetch Selected Document");
		fetchAction.
			setImageDescriptor(PlatformUI.getWorkbench().getSharedImages().
			getImageDescriptor(ISharedImages.IMG_OBJ_FILE));

		showControlPanel =
			new Action("Show Control Panel", IAction.AS_CHECK_BOX) {
				public void run() {
					if (this.isChecked()) {
						layout.topControl = panels[1];
						showControlPanel.setToolTipText("Hide Control Panel");
					} else {
						layout.topControl = panels[0];
						showControlPanel.setToolTipText("Show Control Panel");
						if (autoSearch) {
							launchJob(job);
						}
					}
					parent.layout();
				}
			};
		showControlPanel.setToolTipText("Show Control Panel");
		showControlPanel.setImageDescriptor(SeleneImages.CONTROL_PANEL);
	}

/*	private void showMessage(String message) {
		MessageDialog.openInformation(
			viewer.getControl().getShell(),
			"Selene",
			message);
	} */

	private void hookContextMenu() {
		MenuManager menuMgr = new MenuManager("Selene Menu");
		menuMgr.setRemoveAllWhenShown(true);
		menuMgr.addMenuListener(new IMenuListener() {
			public void menuAboutToShow(IMenuManager manager) {
				SeleneView.this.fillContextMenu(manager);
			}
		});
		Menu menu = menuMgr.createContextMenu(viewer.getControl());
		viewer.getControl().setMenu(menu);
		getSite().registerContextMenu(menuMgr, viewer);
	}

	private void fillContextMenu(IMenuManager manager) {
		manager.add(fetchAction);
		// Other plug-ins can contribute there actions here
		manager.add(new Separator(IWorkbenchActionConstants.MB_ADDITIONS));
	}

	private void hookDoubleClickAction() {
		viewer.addDoubleClickListener(new IDoubleClickListener() {
			public void doubleClick(DoubleClickEvent event) {
				fetchAction.run();
			}
		});
	}

	private void contributeToActionBars() {
		IActionBars bars = getViewSite().getActionBars();
		fillLocalPullDown(bars.getMenuManager());
		fillLocalToolBar(bars.getToolBarManager());
		bars.updateActionBars();
	}

	private void fillLocalPullDown(IMenuManager manager) {
		manager.add(stopSearch);
		manager.add(searchAction);
		manager.add(fetchAction);
		manager.add(showControlPanel);
	}

	private void fillLocalToolBar(IToolBarManager manager) {
		manager.add(stopSearch);
		manager.add(searchAction);
		manager.add(fetchAction);
		manager.add(showControlPanel);
	}

	/**
	 * Passing the focus request to the viewer's control.
	 */
	public void setFocus() {
		viewer.getControl().setFocus();
	}

	public void init(IViewSite site) throws PartInitException {
		super.init(site);
	}
	public void init(IViewSite site, IMemento memento)
		throws PartInitException {
		super.init(site, memento);
		this.memento = memento;
	}
	public void saveState(IMemento memento) {
		super.saveState(memento);
		for (int i = 0; i < MAX_SNIPPETS; i++) {
			if (snippetEditors[i] != null) {
				memento.putString("snippetEditor" + i,
					snippetEditors[i].getEditorInput().getToolTipText());
			}
		}
	}

	public void dispose() {
		getSite().getPage().removePartListener(this);
		for (int i = 0; i < MAX_SNIPPETS; i++) {
			if (snippetEditors[i] != null) {
				page.closeEditor(snippetEditors[i], false);
			}
		}
		super.dispose();
	}

	public void partActivated(IWorkbenchPart part) {
		if (!searchTemporarilyDisabled && part instanceof ITextEditor) {
			for (int i = 0; i < MAX_SNIPPETS; i++) {
				if (snippetEditors[i] == part) {
					searchTemporarilyDisabled = true;
				}
			}
			editor = (ITextEditor)part;
			viewer.setInput(editor.
				getDocumentProvider().getDocument(editor.getEditorInput()));
			searchTemporarilyDisabled = false;
		}
	}
	public void partBroughtToTop(IWorkbenchPart part) {
	}
	public void partClosed(IWorkbenchPart part) {
	}
	public void partDeactivated(IWorkbenchPart part) {
	}
	public void partOpened(IWorkbenchPart part) {
	}
}
