package jp.ne.xuni.selene.views;

import org.eclipse.jface.dialogs.IDialogSettings;

import jp.ne.xuni.selene.SelenePlugin;
import jp.ne.xuni.selene.evaluators.SearchEvaluator;
import jp.ne.xuni.selene.evaluators.SimpleEvaluator;

public class TimeCounter {
	private static final String DS_ID = ".timeCounter";
	private static final int DEFAULT_WAIT = 0;

	private long startTime;
	private int[] timeRecord;
	private int numOfRecords;
	private int counter;

	private SeleneView view;
	private SearchEvaluator eval;
	private boolean printRecord;
	private boolean autoTest;
	private int waitTime = DEFAULT_WAIT;

	private IDialogSettings ds = SelenePlugin.getDefault().
		getDialogSettings().getSection(SelenePlugin.PLUGIN_ID + DS_ID);
	{
		if (ds == null) {
			ds = SelenePlugin.getDefault().getDialogSettings().addNewSection(
				SelenePlugin.PLUGIN_ID + DS_ID);
		}
		printRecord = ds.getBoolean("printRecord");
		autoTest = ds.getBoolean("autoTest");
		try {
			waitTime = ds.getInt("waitTime");
		} catch (NumberFormatException e) { }
	}

	public TimeCounter(SeleneView view) {
		this.view = view;

		String evalClass = ds.get("evalClass");
		if (evalClass != null) {
			try {
				eval = (SearchEvaluator)Class.forName(evalClass).
					getDeclaredConstructor(SeleneView.class).
					newInstance(view);
			} catch (Exception e) {
				e.printStackTrace();
			}
		} else {
			eval = new SimpleEvaluator(view);
		}
	}

	public void setStartTime() {
		startTime = System.currentTimeMillis();
		numOfRecords = 1 + view.numOfSnippets * 3;
		timeRecord = new int[numOfRecords * 2];
		counter = 0;
	}

	public void recordTime() {
		recordTime(-1);
	}
	public synchronized void recordTime(int id) {
		id = (id == -1) ? 0 : id + 1;
		if (counter < numOfRecords) {
			timeRecord[counter * 2] = id;
			timeRecord[counter * 2 + 1] =
				(int)(System.currentTimeMillis() - startTime);
			if (counter++ == numOfRecords - 1) {
				String record = "";
				for (int time : timeRecord) {
					record += time + " ";
				}
				if (printRecord) System.out.println(record);

				eval.evaluate(view.getDocuments());

				if (autoTest) view.timerExec(waitTime, eval);
			}
		}
	}
}
